
function blackBar(marvinImage, x, y, w, h, bar_color) {
    marvinImage.fillRect(
        x, y,
        w, h,
        bar_color//0xFFFFFF00//=0xAARRGGBB
    )
    return marvinImage;
}

function gauss_boxes(sigma, n)  // standard deviation, number of boxes
{
    const wIdeal = Math.sqrt((12 * sigma * sigma / n) + 1);  // Ideal averaging filter width
    let wl = Math.floor(wIdeal);
    if (wl % 2 === 0) wl--;
    const wu = wl + 2;

    const mIdeal = (12 * sigma * sigma - n * wl * wl - 4 * n * wl - 3 * n) / (-4 * wl - 4);
    let m = Math.round(mIdeal);
    // var sigmaActual = Math.sqrt( (m*wl*wl + (n-m)*wu*wu - n)/12 );

    const sizes = [];
    for (let i = 0; i < n; i++) sizes.push(i < m ? wl : wu);
    return sizes;
}

function boxBlurH_3(scl, tcl, w, h, r) {
    for (let i = 0; i < h; i++) {
        for (let j = 0; j < w; j++) {
            let val = 0;
            for (let ix = j - r; ix < j + r + 1; ix++) {
                let x = Math.min(w - 1, Math.max(0, ix));
                val += scl[i * w + x];
            }
            tcl[i * w + j] = val / (r + r + 1);
        }
    }
}

function boxBlurT_3(scl, tcl, w, h, r) {
    for (let i = 0; i < h; i++) {
        for (let j = 0; j < w; j++) {
            let val = 0;
            for (let iy = i - r; iy < i + r + 1; iy++) {
                let y = Math.min(h - 1, Math.max(0, iy));
                val += scl[y * w + j];
            }
            tcl[i * w + j] = val / (r + r + 1);
        }
    }
}

function box_blur_3(scl, tcl, w, h, r) {
    for (let i = 0; i < scl.length; i++) {
        tcl[i] = scl[i];
    }
    boxBlurH_3(tcl, scl, w, h, r);
    boxBlurT_3(scl, tcl, w, h, r);
}

function gauss4(scl, tcl, w, h, r) {
    let bxs = gauss_boxes(r, 3);
    box_blur_3(scl, tcl, w, h, (bxs[0] - 1) / 2);
    box_blur_3(tcl, scl, w, h, (bxs[1] - 1) / 2);
    box_blur_3(scl, tcl, w, h, (bxs[2] - 1) / 2);
}

function image_section_to_source_channels(marvinImage, x, y, w, h) {
    let scl_a = [w * h];
    let scl_r = [w * h];
    let scl_g = [w * h];
    let scl_b = [w * h];
    for (let sect_y = 0; sect_y < h; sect_y++) {
        let glob_y = y + sect_y;
        for (let sect_x = 0; sect_x < w; sect_x++) {
            let glob_x = x + sect_x;
            let index = sect_y * w + sect_x;
            scl_a[index] = marvinImage.getAlphaComponent(glob_x, glob_y);
            scl_r[index] = marvinImage.getIntComponent0(glob_x, glob_y);
            scl_g[index] = marvinImage.getIntComponent1(glob_x, glob_y);
            scl_b[index] = marvinImage.getIntComponent2(glob_x, glob_y);
        }
    }
    return {scl_a, scl_r, scl_g, scl_b};
}

function linear_gaussian(scl_a, scl_r, scl_g, scl_b, w, h, r) {
    let tcl_a = [w * h];
    let tcl_r = [w * h];
    let tcl_g = [w * h];
    let tcl_b = [w * h];

    gauss4(scl_a, tcl_a, w, h, r);
    gauss4(scl_r, tcl_r, w, h, r);
    gauss4(scl_g, tcl_g, w, h, r);
    gauss4(scl_b, tcl_b, w, h, r);
    return {tcl_a, tcl_r, tcl_g, tcl_b};
}

function copy_array(source, target) {
    for (let i = 0; i < source.length; i++) {
        target[i] = source[i];
    }
}

function generateBaseBlur(marvinImage, blur_strength){
    marvinImage.update();
    let canvas = document.createElement('canvas');
    canvas.width =  marvinImage.width;
    canvas.height = marvinImage.height;
    const ctx = canvas.getContext('2d');
    ctx.filter = 'blur('+blur_strength+'px)';
    ctx.drawImage(marvinImage.image, 0, 0, marvinImage.width, marvinImage.height);
    marvinImage.blurred = canvas;
}

function generateBaseBlurImproved(marvinImage, blur_strength){
    marvinImage.update();
    let canvas = document.createElement('canvas');
    let pxext = blur_strength*2;
    let pxext2 = pxext*2;
    canvas.width =  marvinImage.width+pxext2;
    canvas.height = marvinImage.height+pxext2;
    const ctx = canvas.getContext('2d');
    ctx.save();
    //Upper left corner
    ctx.translate(pxext,pxext);
    ctx.scale(-1, -1);
    ctx.drawImage(marvinImage.canvas,
        0, 0, pxext, pxext,
        0, 0, pxext, pxext);
    ctx.restore();
    ctx.save();
    //Top
    ctx.translate(0,pxext);
    ctx.scale(1, -1);
    ctx.drawImage(marvinImage.canvas,
        0, 0, marvinImage.width, pxext,
        pxext, 0, marvinImage.width, pxext);
    ctx.restore();
    ctx.save();
    //Upper right corner
    ctx.translate(canvas.width, pxext);
    ctx.scale(-1, -1);
    ctx.drawImage(marvinImage.canvas,
        marvinImage.width-pxext, 0,pxext, pxext,
        0, 0, pxext, pxext);
    ctx.restore();
    ctx.save();
    //Left
    ctx.translate(pxext, 0);
    ctx.scale(-1, 1);
    ctx.drawImage(marvinImage.canvas,
        0, 0, pxext, marvinImage.height,
        0, pxext, pxext, marvinImage.height
        //0, 0, marvinImage.width, marvinImage.height
    );
    ctx.restore();
    ctx.save();
    //Right
    ctx.translate(canvas.width, 0);
    ctx.scale(-1, 1);
    ctx.drawImage(marvinImage.canvas,
        marvinImage.width-pxext, 0, pxext, marvinImage.height,
        0, pxext, pxext, marvinImage.height);
    ctx.restore();
    ctx.save();
    //Bottom left corner
    ctx.translate(pxext, canvas.height);
    ctx.scale(-1, -1);
    ctx.drawImage(marvinImage.canvas,
            0, marvinImage.height-pxext, pxext, pxext,
            0, 0, pxext, pxext);
    ctx.restore();
    ctx.save();
    //Bottom
    ctx.translate(0, canvas.height);
    ctx.scale(1, -1);
    ctx.drawImage(marvinImage.canvas,
        0, marvinImage.height-pxext, marvinImage.width, pxext,
        pxext, 0, marvinImage.width, pxext);
    ctx.restore();
    ctx.save();
    //Bottom right corner
    ctx.translate(canvas.width, canvas.height);
    ctx.scale(-1, -1);
    ctx.drawImage(marvinImage.canvas,
        marvinImage.width-pxext, marvinImage.height-pxext,pxext, pxext,
        0, 0, pxext, pxext);
    ctx.restore();
    ctx.save();
    //Center
    ctx.drawImage(marvinImage.canvas, pxext, pxext, marvinImage.width, marvinImage.height);
    ctx.filter = 'blur('+blur_strength+'px)';
    ctx.drawImage(canvas, 0, 0, canvas.width, canvas.height);
    let blurred = document.createElement('canvas');
    blurred.width =  marvinImage.width;
    blurred.height = marvinImage.height;
    const ctx_blurred = blurred.getContext('2d');

    ctx_blurred.drawImage(canvas,
        pxext, pxext, marvinImage.width, marvinImage.height,
        0, 0, marvinImage.width, marvinImage.height);

    marvinImage.blurred = blurred;
}

function blurRect(marvinImage, x, y, w, h, blur_strength, scale, radius, mask_scale) {
    if(marvinImage.blurred == null){
        generateBaseBlurImproved(marvinImage, blur_strength);
    }
    marvinImage.update();
    let canvas = document.createElement('canvas');
    let wexp = Math.ceil(w*scale);
    let hexp = Math.ceil(h*scale);
    let xcor = Math.ceil(x-(wexp-w)/2);
    let ycor = Math.ceil(y-(hexp-h)/2)
    canvas.width    = wexp;
    canvas.height   = hexp;
    const ctx = canvas.getContext('2d');
    ctx.fillStyle = 'rgba(0,0,0,0)';
    ctx.fillRect(0, 0, wexp, hexp);
    ctx.fillStyle = '#fff';
    let blur_radius_strength =  Math.ceil(Math.min(wexp,hexp)/100 * radius);
    ctx.filter = 'blur('+blur_radius_strength+'px)';
    ctx.beginPath();
    ctx.ellipse(wexp/2, hexp/2,
        w*mask_scale,
        h*mask_scale, 0, 0, 4 * Math.PI);
    ctx.fill();
    ctx.globalCompositeOperation = 'source-in';
    ctx.drawImage(marvinImage.blurred, xcor,ycor,wexp,hexp, 0,0,wexp,hexp);
    marvinImage.ctx.drawImage(canvas, 0,0,wexp,hexp, xcor,ycor,wexp,hexp);
    //marvinImage.ctx.clearRect(0, 0, marvinImage.width,marvinImage.height);
    //marvinImage.ctx.drawImage(marvinImage.blurred, 0,0,marvinImage.width,marvinImage.height);
    marvinImage.imageData = marvinImage.ctx.getImageData(0, 0, marvinImage.width, marvinImage.height);

    return marvinImage;
}

function blurRect_old(marvinImage, x, y, w, h) {

    const feather_factor = 0.1;
    const blur_radius = 10;

    let {scl_a, scl_r, scl_g, scl_b} = image_section_to_source_channels(marvinImage, x, y, w, h);

    let ocl_a = [w * h];
    let ocl_r = [w * h];
    let ocl_g = [w * h];
    let ocl_b = [w * h];
    copy_array(scl_a, ocl_a);
    copy_array(scl_r, ocl_r);
    copy_array(scl_g, ocl_g);
    copy_array(scl_b, ocl_b);

    let {tcl_a, tcl_r, tcl_g, tcl_b} = linear_gaussian(scl_a, scl_r, scl_g, scl_b, w, h, blur_radius);

    const feather_width = feather_factor * w;
    const feather_height = feather_factor * h;

    let dummy = new MarvinImage();
    Marvin.crop(marvinImage, dummy, x, y, w, h);

    for (let sect_y = 0; sect_y < h; sect_y++) {
        const glob_y = y + sect_y;
        const base_index = sect_y * w;

        const partial_min = feather_factor === 0 ? 1 :
            Math.min(sect_y / feather_height, (h - sect_y) / feather_height);

        for (let sect_x = 0; sect_x < w; sect_x++) {
            const glob_x = x + sect_x;

            const index = base_index + sect_x;

            const f_new = feather_factor === 0 ? 1 :
                Math.min(1, partial_min, sect_x / feather_width, (w - sect_x) / feather_width);
            const f_old = 1 - f_new;

            const this_a = f_new * tcl_a[index] + f_old * ocl_a[index];
            // Correct
            const this_r = f_new * tcl_r[index] + f_old * ocl_r[index];
            const this_g = f_new * tcl_g[index] + f_old * ocl_g[index];
            const this_b = f_new * tcl_b[index] + f_old * ocl_b[index];

            marvinImage.setIntColor(glob_x, glob_y, this_a, this_r, this_g, this_b);
            // dummy.setIntColor(sect_x, sect_y, this_a, this_r, this_g, this_b);
        }
    }

    marvinImage.update();
    // dummy.update();
    return marvinImage;
}


function pixelRectGlitch(marvinImage, x, y, w, h, pixel_size = 20, distance = 0) {
    //const pixel_size = 10; // todo: calc from image size / box size / user pref.
    if(pixel_size == null){
        //Choose pixel_size based on with / height
    }
    var boxcolor = marvinImage.getIntColor(x,y);
    var px = x;
    var py = y;
    var count_i = 0;
    var count_j = 0;
    for(var i=x; i<x+w; i++){
        if(count_i%pixel_size == 0){
            px = i;
            boxcolor = calculatePixelBoxColor(marvinImage, px, py, px+pixel_size, py+pixel_size, x, y, w, h, pixel_size);
        }
        for(var j=y; j<y+h; j++){
            if(i < marvinImage.getWidth() && j < marvinImage.getHeight()){
                if(count_j%pixel_size == 0){
                    py = j;
                    boxcolor = calculatePixelBoxColor(marvinImage, px, py, px+pixel_size, py+pixel_size, x, y, w, h, pixel_size);
                }
                marvinImage.setIntColor(i,j,boxcolor);
            }
            count_j++;
        }
        count_i++;
        count_j = 0;
        px = i;
    }
    marvinImage.update();
    return marvinImage;
}

function pixelRect(marvinImage, x, y, w, h, pixel_size = 20, distance = 0) {
    for(let i=x; i<x+w; i+=pixel_size+distance){
        for(let j=y; j<y+h; j+=pixel_size+distance){
            let box_color = calculatePixelBoxColor(marvinImage, i, j, i+pixel_size, j+pixel_size, x, y, w, h, pixel_size);
            //new pixel box begins here
            for(let x1=i; x1<i+pixel_size; x1++) {
                for(let y1=j; y1<j+pixel_size; y1++) {
                    if (x1 < x+w && y1 < y+h) {
                        marvinImage.setIntColor(x1, y1, box_color);
                    }
                }
            }
        }
    }
    marvinImage.update();
    return marvinImage;
}

function pixelRectFast(marvinImage, x, y, w, h, pixel_size = 20, distance = 0) {
    for(let i=x; i<x+w; i+=pixel_size+distance){
        for(let j=y; j<y+h; j+=pixel_size+distance){
            //let box_color = calculatePixelBoxColor(marvinImage, i, j, i+pixel_size, j+pixel_size, x, y, w, h, pixel_size);
            let box_color = marvinImage.getIntColor(i,j);
            //new pixel box begins here
            for(let x1=i; x1<i+pixel_size; x1++) {
                for(let y1=j; y1<j+pixel_size; y1++) {
                    if (x1 < x+w && y1 < y+h) {
                        marvinImage.setIntColor(x1, y1, box_color);
                    }
                }
            }
        }
    }
    marvinImage.update();
    return marvinImage;
}

function drawHexagon(marvinImage, x, y, r, a) {
    marvinImage.ctx.beginPath();
    for (let i = 0; i < 6; i++) {
        marvinImage.ctx.lineTo(x + r * Math.cos(a * i), y + r * Math.sin(a * i));
    }
    marvinImage.ctx.fillStyle = toColor(calculatePixelBoxColor(marvinImage, Math.ceil(x), Math.ceil(y), 0,0,0,0,0,0,r));
    marvinImage.ctx.fill();
    marvinImage.ctx.closePath();
    //marvinImage.ctx.stroke();
}

function toColor(num, alpha = false) {
    num >>>= 0;
    let b = num & 0xFF,
        g = (num & 0xFF00) >>> 8,
        r = (num & 0xFF0000) >>> 16,
        a = ( (num & 0xFF000000) >>> 24 ) / 255 ;
    if(!alpha){
        a = 1;
    }
    return "rgba(" + [r, g, b, a].join(",") + ")";
}

function drawGrid(marvinImage, x0, y0, width, height, size, distance = 0, glitch = false) {
    let a = 2 * Math.PI / 6;
    let r = size;
    if(glitch){
        a = Math.ceil(a);
    }
    for (let y = y0+r; y + r * Math.sin(a) < y0+height; y += 2*(r * Math.sin(a))) {
        y += distance;
        for (let x = x0+r, j = 0; x + r * (1 + Math.cos(a)) < x0+width; x += r * (1 + Math.cos(a))) {
            x += distance;
            let yc = y + 0.5*((-1) ** j++ * r * Math.sin(a));
            drawHexagon(marvinImage, x, yc, r, a);
        }
    }
    marvinImage.imageData = marvinImage.ctx.getImageData(0, 0, marvinImage.width, marvinImage.height);
    marvinImage.update();
    return marvinImage;
}



/**
 *
 * @param marvinImage
 * @param x1 = pixel box starting x
 * @param y1 = pixel box starting y
 * @param x2 = pixel box ending x
 * @param y2 = pixel box ending y
 * @param x = entire rectangle x
 * @param y = entire rectangle y
 * @param w = rectangle width
 * @param h = rectangle height
 * @param pixel_size = box size
 * @returns {*}
 */
function calculatePixelBoxColor(marvinImage, x1, y1, x2, y2, x, y, w, h, pixel_size) {
    switch (1) {
        case 0:
            return marvinImage.getIntColor(x1,y1);
        case 1:
            let rectarea = new MarvinImage();
            let pixelmarvin = new MarvinImage();
            Marvin.crop(marvinImage, rectarea, x1, y1, pixel_size, pixel_size);
            Marvin.scale(rectarea, pixelmarvin, 1, 1);
            return pixelmarvin.getIntColor(0,0);
            break;
        default:
    }
    return marvinImage.getIntColor(x1,y1);
}


