

function paintCensorRequest(marvinImage, processedRequest, bar_configuration, pixel_configuration, blur_configuration) {
    let labels = processedRequest.labels;
    let valid_detections = processedRequest.valid_detections;
    let classes = processedRequest.classes;
    let scores = processedRequest.scores;
    let boxes = processedRequest.boxes;
    //let naturalWidth = processedRequest.naturalWidth;
    //let naturalHeight = processedRequest.naturalHeight;
    let censor_type = processedRequest.censortype;
    paintCensor(marvinImage, censor_type, labels, valid_detections, classes, scores, boxes, bar_configuration, pixel_configuration, blur_configuration);
}

function paintCensor(marvinImage, censor_type, labels, valid_detections, classes, scores, boxes,
                     bar_configuration,
                     pixel_configuration,
                     blur_configuration
                     ){
    let naturalWidth = marvinImage.canvas.width;
    let naturalHeight = marvinImage.canvas.height;

    let censor_type_numeral = (censor_type == "black"? 3 : (censor_type == "pixel"? 1 : (censor_type == "blur"? 2 : (censor_type == "box"? 4 : 0))));

    log(
        "valid detections", valid_detections,
        "classes",classes,
        "boxes",boxes,
        "scores",scores);

    for (let i = 0; i < valid_detections[0]; ++i) {
        const klass = names[classes[i]];
        if(labels.includes(klass)) {
            let [x1, y1, x2, y2] = boxes.slice(i * 4, (i + 1) * 4);
            const score = scores[i].toFixed(2);
            x1 *= naturalWidth;
            x2 *= naturalWidth;
            y1 *= naturalHeight;
            y2 *= naturalHeight;
            let x = Math.ceil(x1);
            let y = Math.ceil(y1);
            let w = Math.ceil(x2 - x1);
            let h = Math.ceil(y2 - y1);
            switch (censor_type_numeral) {
                case 0:
                    break;
                case 1:
                    let wexp = Math.ceil(w*pixel_configuration._scale);
                    let hexp = Math.ceil(h*pixel_configuration._scale);
                    let xcor = Math.ceil(x-(wexp-w)/2);
                    let ycor = Math.ceil(y-(hexp-h)/2);
                    let distance = pixel_configuration._distance;
                    let box_size = pixel_configuration._pixel_size;
                    if(pixel_configuration._pixel_size_mode == 0){
                        //box_size = Math.ceil(Math.sqrt(naturalHeight*naturalWidth)/100*pixel_configuration._pixel_size);
                        box_size = Math.ceil(Math.min(naturalHeight,naturalWidth)/100*pixel_configuration._pixel_size);
                    }

                    if(pixel_configuration._geometry == 1){
                        marvinImage = drawGrid(marvinImage, xcor, ycor, wexp, hexp, box_size, distance);
                    }else if(pixel_configuration._geometry == 2){
                        marvinImage = drawGrid(marvinImage, xcor, ycor, wexp, hexp, box_size, distance, true);
                    }else{
                        if(box_size < 5){
                            marvinImage = pixelRectFast(marvinImage, xcor, ycor, wexp, hexp, box_size, distance);
                        }else{
                            marvinImage = pixelRectFast(marvinImage, xcor, ycor, wexp, hexp, box_size, distance);
                        }
                    }
                    break;
                case 2:
                    marvinImage = blurRect(marvinImage, x, y, w, h,
                        blur_configuration._blur_strength,
                        blur_configuration._scale,
                        blur_configuration._blur_radius,
                        blur_configuration._mask_scale);
                    break;
                case 3:
                    marvinImage = blackBar(marvinImage, x, y, w, h, bar_configuration._color);
                    break;
                case 4:
                    const font = "16px sans-serif";
                    marvinImage.ctx.font = font;
                    marvinImage.ctx.textBaseline = "top";
                    marvinImage.ctx.strokeStyle = "#00FFFF";
                    marvinImage.ctx.lineWidth = 4;
                    marvinImage.ctx.strokeRect(x, y, w, h);
                    marvinImage.ctx.fillStyle = "#00FFFF";
                    const textWidth = marvinImage.ctx.measureText(klass + ":" + score).width;
                    const textHeight = parseInt(font, 10); // base 10
                    marvinImage.ctx.fillRect(x, y, textWidth + 4, textHeight + 4);
                    marvinImage.ctx.fillStyle = "#000000";
                    marvinImage.ctx.fillText(klass + ":" + score, x1, y1);
                    marvinImage.imageData = marvinImage.ctx.getImageData(0, 0, naturalWidth, naturalHeight);
                    break;
                //default:

            }
        }
    }
}

